
/* MPIPairwiseStatSig2.0 : A Program for estimating normal and/or non-conservative 
pairwise statistical significance of a sequence alignment using MPI.

Please cite:

Ankit Agrawal, Sanchit Misra, Daniel Honbo, and Alok Choudhary, "Parallel 
Pairwise Statistical Significance of Local Sequence Alignment Using 
MPI Library", under review, 2010. 

Copyright (c) 2010 Ankit Agrawal
All Rights Reserved.  E-mail: ankitag@eecs.northwestern.edu

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice, this paragraph and the following three paragraphs
appear in all copies.

Permission to incorporate this software into commercial products may be
obtained from Iowa State University Research Foundation, Inc.

	      Ankit Agrawal
	      Department of Computer Science
	      Iowa State University
	      Ames, IA 50011
	      E-mail: ankitag@cs.iastate.edu
	      WWW: http://www.cs.iastate.edu/~ankitag

    Proper attribution of the author as the source of the software would
    be appreciated:

    The MPIPairwiseStatSig program estimates the normal and/or non-conservative pairwise statistical significance of a pairwise local 
    alignment of two protein sequences independent of any database using censored Maximum Likelihood Fitting.

    The current version also supports the use of sequence-specific and position-specific substitution substitution matrices. 

    The MPIPairwiseStatSig program is written in C using MPI library and runs under Unix and Linux systems.
    We think that the program is portable to many machines.

    The pair of sequences to be compared are stored in a single file in FASTA format.
    The first line begins with the symbol '>'
    followed by the name of the sequence. The sequence is on the remaining lines.
    
>Human-beta
VHLTPEEKSAVTALWGKVNVDEVGGEALGRLLVVYPWTQRFFESFGDLSTPDAVMGNPKV
KAHGKKVLGAFSDGLAHLDNLKGTFATLSELHCDKLHVDPENFRLLGNVLVCVLAHHFGK
EFTPPVQAAYQKVVAGVANALAHKYH


  Compiling:
		
  mpicc MPIPairwiseStatSig.c gap3PSSM.comp.c histogram.c wisememman.c  -o MPIPairwiseStatSig -lm 
   
  Running:
	 ./MPIPairwiseStatSig <sequencePairFilename> [options] 
	 
	 SSM: Standard substitution matrix
	 SSSM: Sequence-specific substitution matrix
	 PSSM: Position-specific substitution matrix
	 
	revised by Yuhong Zhang
	06/08/2011

*/
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <Histogram.h>
#include <MPIPairwiseStatSig.h>
#include <gap3PSSM.comp.h>
#include <generic.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <set>
#include <iostream>
using namespace std;
#define NEED_PSSE 0
int StatSig::BlmMatrix[128][128] 	= {{0, 0}};

StatSig::StatSig(CParams* params)
{
	totalTime 	= 0;
	IOTime 		= 0;
	alnTime 	= 0;
	gatherTime 	= 0;
	fitTime 	= 0;
	
	lumpsize 	= 20;
	lowscore 	= 0;	
	highscore 	= 5000; 
	no_entry 	= 0;
	
	MatrixFilePath 	= NULL;
	DBFile 			= NULL;
	seqQueryFile 	= NULL;
	seqClsFile 		= NULL; 
	dbSeqClf 		= NULL;
	
	
	PSSMatrix		= NULL;
	LocalMatrix   	= NULL;

	
	amino_acids			= amino_acids_23;
	amino_acids_trans 	= amino_acids_trans_23;
	
	this->params = params;
}
StatSig::~StatSig()
{
	if( DBSeqs )
	{
		free(DBSeqs);
		DBSeqs = NULL;
	}
	
	if ( QuerySeqs )
	{
		free(QuerySeqs);
		QuerySeqs = NULL;
	}
	
	if ( dbSeqClf )
	{
		for(int i = 0 ; i < TotalSeqs ; i++)
		{
			free (dbSeqClf[i]);			
		}
		free(dbSeqClf);
		dbSeqClf = NULL;
	}
	
	
}

void StatSig::run()
{
	
	double totalTimeStart = MPI::Wtime();
	
	nProcessors = MPI::COMM_WORLD.Get_size(); /* No. of processors */
	pRank 		= MPI::COMM_WORLD.Get_rank(); /* rank of processor */

	// read parameters from users' input parameters
	Etype = params->getEstimateType();
	Mtype = params->getMatrixType();
	MatrixFilePath = params->getMatrixFilePath();

	if (Mtype == 1){ /* HANDLING path which include PSSM file */
		if ( strlen(MatrixFilePath) <= 1 ){
			if(pRank == 0){
				printf("Invalid path for PSSM files\n");
			}
			MPI::Finalize();
			return;
			
		}else if (MatrixFilePath[strlen(MatrixFilePath) - 1] == '/'){		
			//delete the the lastest character '/' if it exits
			MatrixFilePath[strlen(MatrixFilePath) -1] = '\0';
		}
		
        gapOpen = 11; // PSI-BLAST default
		gapExten = 1; // PSI-BLAST default
	} else { /*loading substitution matrix */
		

		double time_s = MPI::Wtime();
		getMatrixFromFile (MatrixFilePath, BlmMatrix);

		IOTime += ( MPI::Wtime() - time_s);
        
        gapOpen = 11; // BL62 best [Agr08]
		gapExten = 2; //BL62 best [Agr08]	
		
	}
	

	DBFile 			= params->getDBFile();	
	seqQueryFile	= params->getQueryFile();
	seqClsFile 		= params->getClsFile();
	
	shuffles_ML = params->getShuffle();	

	// read sequence pairs
	double seqFileReadStart = MPI::Wtime();

  	ReadSeqs();

	double seqFileReadTimeL = MPI::Wtime() - seqFileReadStart;
	
	IOTime += seqFileReadTimeL ;
						
	// call MPI-PSS function
	GetPSSE();

	double totalTimeEnd = MPI::Wtime();
	
	totalTime = totalTimeEnd - totalTimeStart;
	
}

void StatSig::GetPSSE()
{

	int queryLen, subLen;
	char* query;
	char* subject;
	char* queryName;
	char* subName;
#if NEED_PSSE
	char PvalueScvFileName[MAX_PATH_LENGTH];
	char PValueRltPath[MAX_PATH_LENGTH];
	CompRec** PValueInfo 	= NULL;
	CompRec* PSSInfo 		= NULL;	
	FILE* pEPQVC;

	if(pRank == 0){

		iniStr(PValueRltPath, MAX_PATH_LENGTH);
		sprintf(PValueRltPath,"SortPvalueScv-mpi%d", nProcessors);
		mkdir(PValueRltPath, 0777);	

		// allocate memory for PSSE information ,then resorted and write them to .csv files.
		PValueInfo = (CompRec**) malloc(sizeof(CompRec*) * TotalQuery );
		if (PValueInfo == NULL){
			ERROR_OUT("Error: no memory space for database sequences \n");
		}

		for (int i = 0; i < TotalQuery; i++){
			PValueInfo[i] =  (CompRec*) calloc( TotalSeqs , sizeof(CompRec));
			if (PValueInfo[i] == NULL){
				ERROR_OUT("Error: no memory space for database sequences \n");
			}
		}
		
		char allQueryStat[MAX_PATH_LENGTH];
		iniStr(allQueryStat, MAX_PATH_LENGTH);
		sprintf(allQueryStat, "./%s/All_%d_query_%d_lib_EPQVC.csv\0",PValueRltPath,
					TotalQuery, TotalSeqs);
		
		pEPQVC= fopen(allQueryStat,"w");
		if (pEPQVC == NULL){
			ERROR_OUT ("Error : cannot open file : %s\n");
			return ;
		}
		// write the captain row
		fprintf(pEPQVC, "QueryNo,QueryName,True Positives,Coverage at certain Number of false positives\n");
		fprintf(pEPQVC, " , , ,");
		for (int i = 1 ; i <= 100; i++){
			fprintf(pEPQVC, "%d,", i);
        }
		fprintf(pEPQVC, "\n");
	
		// allocate memory for the PSSE information each pair		
		PSSInfo = (CompRec*) malloc(sizeof(CompRec) * TotalSeqs);
		if (PSSInfo == NULL){
			ERROR_OUT("Error: no memory space for database sequences \n");
		}
	}
#endif

	    	
	for (int qindex = 0; qindex < TotalQuery; qindex++)
	{		
#if NEED_PSSE			
		iniStr(PvalueScvFileName, MAX_PATH_LENGTH);
		sprintf(PvalueScvFileName, "./%s/No_%d_%s_query_%d_lib_%d.csv\0", 
				PValueRltPath, qindex + 1, QuerySeqs[qindex].name, TotalQuery, TotalSeqs);
#endif		
		queryLen 	= QuerySeqs[qindex].length;
		query 		= QuerySeqs[qindex].context;
        queryName 	= QuerySeqs[qindex].name;

		//for (int sindex = 0; sindex < 2; sindex++)
		for (int sindex = 0; sindex < TotalSeqs; sindex++)
		{			
			subLen 	= DBSeqs[sindex].length;
			subject = DBSeqs[sindex].context;
			subName = DBSeqs[sindex].name;

			
			StatSigForSeqPair_ML(query, queryName, queryLen, subject, subName, subLen);
			
#if NEED_PSSE
			if(pRank == 0){
				PSSInfo[sindex].seq1 = qindex + 1; // number index start from 1
				iniStr(PSSInfo[sindex].name1, MaxNameLen);
				memmove (PSSInfo[sindex].name1, QuerySeqs[qindex].name, strlen(QuerySeqs[qindex].name));
				
				PSSInfo[sindex].seq2 = sindex + 1; // number index start from 1
				iniStr(PSSInfo[sindex].name2, MaxNameLen);
				memmove (PSSInfo[sindex].name2, DBSeqs[sindex].name, strlen(DBSeqs[sindex].name));
				
				if (QuerySeqs[qindex].cls[0] != DBSeqs[sindex].cls[0] || QuerySeqs[qindex].cls[1] != DBSeqs[sindex].cls[1])
				{
					PSSInfo[sindex].topolog = 0;
					PSSInfo[sindex].homolog = 0;
					
				} else if (QuerySeqs[qindex].cls[2] == DBSeqs[sindex].cls[2]) {
					PSSInfo[sindex].topolog = 1;
					if (QuerySeqs[qindex].cls[3] == DBSeqs[sindex].cls[3])
					{
						PSSInfo[sindex].homolog = 1;
						
					} else {
						PSSInfo[sindex].homolog = 0;
						
					}
				} else {
					PSSInfo[sindex].topolog = 0;
					PSSInfo[sindex].homolog = 0;					
				}
				// just for testing, you can delete it!
				#if 0				
				 cout << "------------------------------"<< endl;
				 cout << "lambda = " << lambda_ML_censor [0] << endl;
				 cout << "K = " << K_ML_censor[0] << endl;
				 cout << "score = "<< pwAlgnScore[0] << endl;
				 cout << "Pvalue = " << pwStatSig[0] << endl;
				 cout << "------------------------------"<< endl;								
				#endif
				
				if(Etype == 0)  { /* normal */
					PSSInfo[sindex].lambda 	= lambda_ML_censor [0];
					PSSInfo[sindex].K 		= K_ML_censor[0];
					PSSInfo[sindex].score 	= pwAlgnScore[0];
					PSSInfo[sindex].Pvalue 	= pwStatSig[0];
				} else if (Etype == 1) { /* min PSSE */
					int p =0;
					if (pwStatSig[0] > pwStatSig[1]) p = 1;
					PSSInfo[sindex].lambda 	= lambda_ML_censor [p];
					PSSInfo[sindex].K 		= K_ML_censor[p];
					PSSInfo[sindex].score 	= pwAlgnScore[p];
					PSSInfo[sindex].Pvalue 	= pwStatSig[p];

				} else if (Etype == 2 ) { /* max PSSE */
					int p =0;
					if (pwStatSig[0] < pwStatSig[1]) p = 1;
					PSSInfo[sindex].lambda 	= lambda_ML_censor [p];
					PSSInfo[sindex].K 		= K_ML_censor[p];
					PSSInfo[sindex].score 	= pwAlgnScore[p];
					PSSInfo[sindex].Pvalue 	= pwStatSig[p] ;
				} else { /* average PSSE */
					PSSInfo[sindex].lambda 	= lambda_ML_censor [0];
					PSSInfo[sindex].K 		= K_ML_censor[0];
					PSSInfo[sindex].score 	= pwAlgnScore[0];
					PSSInfo[sindex].Pvalue 	= ( pwStatSig[0] +  pwStatSig[1]) / 2;
				}							
			}
			
#endif	               			
		} //end for sindex
		
		
#if NEED_PSSE		
		if(pRank == 0)
		{
			//using quick sort to sort the pssm information			
			qsort(PSSInfo, TotalSeqs, sizeof(CompRec), compar_ascent_Pvalue);								
			// compute EPQ information		
			int homologCount = 0;
			double* coverage = (double*) calloc(101, sizeof(double));
			if (coverage == NULL){
				ERROR_OUT("Error: no memory space for database sequences \n");
			}

			for (int pp = 0 ; pp < TotalSeqs; pp ++)
			{				
				homologCount += PSSInfo[ pp ].homolog;
			}

			int FPNum 	= 0; 	// Number of False positives
			int TPNum 	= 0;	// Number of True Positives
			int temp	= 0;
			
			
			if (homologCount > 0){
				
				while(FPNum < 100){
					if (PSSInfo[temp].homolog == 1){
						TPNum ++;					
					} else {
						FPNum ++;
					}
					coverage[ FPNum ] = (double)TPNum / homologCount;
					temp ++;
				}

				fprintf(pEPQVC, "%d,%s,%d,", qindex + 1, QuerySeqs[ qindex ].name, homologCount );
				for (int pp = 0; pp < 100; pp++)
				{
					fprintf(pEPQVC, "%e,",coverage[pp]);	
				}
				fprintf(pEPQVC, "\n");
			}
		
			// free PSSInfo memory after write into .csv files
			if(coverage){
				free(coverage);
				coverage = NULL;
			}
			
			double write_time_start = getSysTime();
	
			writePSSToFile(PSSInfo, TotalSeqs, PvalueScvFileName);

			double write_time_end = getSysTime();

			IOTime += (write_time_end - write_time_start);
			
		}
	#endif	
	}// end for qindex

#if NEED_PSSE	
	if(pRank == 0){
	// release the memory for each pair-wise sequence PSSE information
		if (PSSInfo){
			free(PSSInfo);
			PSSInfo = NULL;
		}
		fclose( pEPQVC );
	}
#endif

}


void StatSig::StatSigForSeqPair_ML(char* query, char* queryName, int queryLen,
								   char* subject, char* subName, int subLen)
{ 

	
	double 	SSMReadTimeStart, SSMReadTimeL,			
			alnStart, alnTimeL,			
			gatherStart, gatherTimeL,			
			fitTimeStart, fitTimeL;
			
			
	srand(time(NULL) + pRank * pRank);		
	
	char* subsMatFile = NULL;
	
  	 			
	float 	ret_mu = 0, 
			ret_lambda = 0, 
			ret_Kmn;
			
	
	char* seq1 = (char*)calloc(queryLen + 2, sizeof(char)); // +2 for the ending and head character '#'
	char* seq2 = (char*)calloc(subLen + 2, sizeof(char));
	
	seq1[0] 			= '#';	
	memmove (seq1 + 1, query, queryLen);
	seq1[queryLen + 1] 	= '#';
	
	
	seq2[0] 			= '#';
	memmove (seq2 + 1, subject, subLen);
	seq2[subLen + 1] 	= '#';
	

	int len1Ptr 	= queryLen; 
	int	len2Ptr 	= subLen;
	char* seq1orig	= query;
	char* seq2orig	= subject;
	
	char* seq1Ptr 	= seq1; 
	char* seq2Ptr 	= seq2;
	
	shuffles_ML = ceil(1.0 * shuffles_ML / nProcessors) * nProcessors;
	int n_shufflesP = shuffles_ML / nProcessors;
	
	int scores[shuffles_ML];
	int scoresP[n_shufflesP];
	
	int ii = 0; 	
	char* sflSeq 		= NULL;
	char* tempPSSMName 	= queryName;
	CHistogram* His 	= NULL;
	
startAlgn:


	if (ii == 1) {
		// swap seq1 and seq2 information including sequence and length.
		seq1Ptr = seq2; 
		seq2Ptr = seq1;

		seq1orig	= subject;
		seq2orig	= query;
		
		len1Ptr = subLen; 
		len2Ptr = queryLen;
		tempPSSMName = subName;
	}
	

	His = new CHistogram;	
	if (pRank == 0){		
		His->new_Histogram(lowscore, highscore, lumpsize);
	}

	SSMReadTimeStart = MPI::Wtime(); 
	
	if (Mtype == 1) {
		/* loading pssm matrix from file */
        iniStr(MatFilename, MAX_PATH_LENGTH);
		
		sprintf(MatFilename, "./%s/pssm_nr_%s_iter5.txt\0",
				MatrixFilePath, tempPSSMName);
		
		GetPSSM(MatFilename, gapOpen, gapExten, 99999, len1Ptr, len2Ptr);	
		
	} else {
		GetSM(MatrixFilePath, gapOpen, gapExten, 99999, len1Ptr, len2Ptr);
		SM2PSSM(seq1orig,len1Ptr);
	}
	
	SSMReadTimeL = MPI::Wtime() - SSMReadTimeStart; 
 	IOTime += SSMReadTimeL; 
	
	int score = 0, j = 0;	
	alnStart = MPI::Wtime(); //start counting total time

	if (pRank == 0) {
		scoresP[0] = score = UseGAP3(seq1Ptr, seq2Ptr, len1Ptr, len2Ptr);
		//CHECK(score)					 
		pwAlgnScore[ii] = (double)score;
		j = 1;
	}
	if (sflSeq)
	{
		free(sflSeq);
		sflSeq = NULL;
	} 
	
	sflSeq = (char*)malloc( (len2Ptr + 2 )* sizeof (char));
	if (sflSeq == NULL)
	{
		ERROR_OUT("Error: no memory space for database sequences \n");		
	}			

	
	//copy original subject sequences for shuffling
   
	sflSeq[0] 				= '#';
	sflSeq[len2Ptr + 1] 	= '#';
	
	for (; j < n_shufflesP; j++) 
	{	
		memmove(sflSeq + 1, seq2orig, len2Ptr);
		ShuffleSeq(sflSeq,len2Ptr);

		scoresP[j] = UseGAP3(seq1Ptr, sflSeq, len1Ptr, len2Ptr);

	}
	
	alnTimeL = MPI::Wtime() - alnStart;
	alnTime += alnTimeL;               //end counting computation time

	gatherStart = MPI::Wtime();   	
			
	MPI::COMM_WORLD.Gather(scoresP, n_shufflesP, MPI::INT, scores, n_shufflesP, MPI::INT, 0);

	gatherTimeL = MPI::Wtime() - gatherStart;;
	gatherTime += gatherTimeL;  
	
	int gap4maxScore = 0;
	
	if( pRank == 0 ) 
	{
		
		fitTimeStart = MPI::Wtime();
		
		for(j = 0; j < shuffles_ML; j++) 
		{
			score = scores[j];
			His->AddToHistogram(score);
			//CHECK(score)
			if ( score > highscore )  score = highscore;
	   		if (score > gap4maxScore) gap4maxScore = score;

		}
		
		/* This code is for ML fitting*/
		/*
		Censoring left of peak with automatic estimation of nBelowCutoff
		*/
	
		ret_mu = 0, ret_lambda = 0, ret_Kmn;
		His->ExtremeValueFitHistogram(CENSOR_LEFT_OF_PEAK, 0, 0, 
								gap4maxScore, &ret_mu, &ret_lambda);
		
		
		ret_Kmn = exp((ret_mu) * (ret_lambda));
		lambda_ML_censor[ii] = (double)ret_lambda;
		K_ML_censor[ii] = (double)ret_Kmn / (queryLen * subLen);
		
		
		pwStatSig[ii] = 1 - exp(-ret_Kmn * exp(-ret_lambda*  (pwAlgnScore[ii])));
		/*	ML fitting code ends */
		fitTimeL = MPI::Wtime() - fitTimeStart;
		fitTime += fitTimeL ; 
		
		delete His;
		
	}

	if ((Etype > 0) && ii==0) {
		FreeArrays();
		ii = 1;		
		goto startAlgn;
	}
	
	FreeArrays();
	
	if (sflSeq){ free(sflSeq); sflSeq = NULL;} 
	
	if (seq1) {free(seq1); seq1 = NULL; }
	if (seq2) {free(seq2); seq2 = NULL; }

}

#define INIT_DB_SIZE	3000
#define INIT_QUERY_SIZE	100

void StatSig::ReadSeqs()
{		

	double time_start, time_end, time_last;
	time_start = getSysTime();
	
	int numSeqs 	= -1;
	int numQuery 	= 0;
	int nameLen 	= 0;
	int contextLen 	= 0;
	int init_size 	= INIT_DB_SIZE;
	
	char* pName 	= NULL;
	char* pContext 	= NULL;
	
	DBSeqs = (SeqInfo* )malloc ( init_size * sizeof (SeqInfo));
	
	if (DBSeqs == NULL)
	{
		 ERROR_OUT("Error: no memory space for database sequences \n");	
	}
	
	fp_DB = fopen (DBFile, "r");
	if (!fp_DB)
	{
		cout << "Error: opening file '"
			 << DBFile <<"' fails!"<<endl;
		exit(0);
	}

	
	char ch;			
	ch = getc(fp_DB);
	while(ch != EOF) 
	{
		if (numSeqs >= INIT_DB_SIZE)
		{
			init_size *= 2;
			DBSeqs = (SeqInfo* )realloc (DBSeqs, init_size * sizeof (SeqInfo));
			if (DBSeqs == NULL)
			{
				 ERROR_OUT("Error: no memory space for database sequences \n");	
			}
		}
		
		if (ch == '>')
		{		
			numSeqs++;
			iniStr(DBSeqs[numSeqs].context, MaxSeqLen);
			iniStr(DBSeqs[numSeqs].name, MaxNameLen);

			pName = DBSeqs[numSeqs].name;
			ch = getc(fp_DB);
			
			nameLen = 0;
			pName[nameLen] = ch;
						
			// the length of whole name is fixed as 6 for CATH database
			while(++nameLen < 6) {
				ch = getc(fp_DB);
				pName[nameLen] = ch;
			}			
			
			// skip the classification of sequence
			while (ch != '\n')
				ch = fgetc(fp_DB);
				
			pContext = DBSeqs[numSeqs].context;
			
			contextLen = 0;
			// read the sequence characters
			while(ch != '>' && ch != EOF) {	
				// filter invalid characters
				if(ch != '?' && ch != '\n' && ch != '\r' && ch != ' '){
					pContext[ contextLen++ ] = ch;
				}
				ch = getc(fp_DB);
			}
			
			DBSeqs[numSeqs].length = contextLen;
		}else{
		
			ch = getc(fp_DB);
		}
		
	}

	if(numSeqs == -1){
		cout << "Error : cannot find FASTA famat from dabase of subject sequences!" <<endl;
		exit(0);
	}
	TotalSeqs = ++numSeqs; 
	
	// read classification file
	loadClsOfSeqs();

	for (int kk = 0; kk < TotalSeqs; kk ++)
	{
		for (int lev = 0; lev < 4; lev ++)
		{
			DBSeqs[kk].cls[lev] = dbSeqClf[kk][lev];
		}
	}


	/* 
		handle the queries name, read them first, then 
		find queries indices in the dababase of subject sequence
	*/
	int index;
	fp_querySeq = fopen (seqQueryFile , "r");	
	if (!fp_querySeq)
	{
		cout << "Error: opening file '" 
			 << seqQueryFile <<"' fails!"<<endl;
		exit(0);
	}
	
	int ini_query_size = INIT_QUERY_SIZE;
	
	QuerySeqs = (SeqInfo* )malloc ( ini_query_size * sizeof (SeqInfo));
	if (QuerySeqs == NULL)
	{
		 ERROR_OUT("Error: no memory space for database sequences \n");	
	}
	
	char tempName[MaxNameLen];
	ch = getc(fp_querySeq);
	while(ch != EOF) 
	{
		if (numQuery >= INIT_QUERY_SIZE) {
			ini_query_size *= 2;
			QuerySeqs = (SeqInfo* )realloc (QuerySeqs,  ini_query_size * sizeof (SeqInfo));
			if (QuerySeqs == NULL)
			{
				 ERROR_OUT("Error: no memory space for database sequences \n");	
			}
		}
		
		while(ch == '\n' || ch == '\r') {
			ch = getc(fp_querySeq);
		}

		iniStr(tempName,MaxNameLen);
		int i = 0;
		
		//read one query name
		while(ch != '\r' && ch !='\n' && ch != EOF) {
			tempName[i++] = ch;
			ch = getc(fp_querySeq);
		} //tempName now contains a new sequence name		
		
		
		index = FindStringIndex(tempName, DBSeqs, TotalSeqs);
				
		if (index == -1) {
			cout <<"Error: Query "
				 << tempName 
				 <<" not found in the database!\n";
			exit(0);
		
		}else{ // if found it from dababase , copy its information into query sequence
			
			strcpy(QuerySeqs[numQuery].name, DBSeqs[index].name);
			strcpy(QuerySeqs[numQuery].context, DBSeqs[index].context);
			QuerySeqs[numQuery].length = DBSeqs[index].length;
			
			for(int cc = 0; cc < 4; cc++){
				QuerySeqs[numQuery].cls[cc] = dbSeqClf[index][cc];
			}
			
			numQuery++;
		}
		
		if(ch == EOF) break;					
		ch = getc(fp_querySeq);				//all this is done to handle \r\n
		if(ch == '\n'){
			ch = getc(fp_querySeq);
		}
	}
	
	TotalQuery = numQuery; //numQuery is the number of query sequences
	
	time_end = getSysTime();
	
	time_last = time_start - time_end;
	
	IOTime += time_last;


}




/* Produce a random sequence of given length and composition */
void StatSig::ShuffleSeq(char seq[], int len)
{ 
	long  j, k;		/* index */
	char  tmp;		/* tmp base */
	int   one;

    for ( one = 0; one < 2; one++ )
    { 
		for ( j = 1; j <= len; j++ )
		{ 
			k = (long)(rand()/(RAND_MAX+1.0)*len)+1;
			/*k = k % len + 1;*/
			if ( j != k )
			{ 
				tmp = seq[j];
				seq[j] = seq[k];
				seq[k] = tmp;
			}
		}
    }
}


void StatSig::PrintResult()
{	
	printf("\nTime (MPI_Wtime) = %f \r\n",nProcessors, totalTime);

	printf("nProcessors, shuffles, IOTime, alnTime, gatherTime, fitTime\n");
	printf("%d, %d, %f, %f, %f, %f \n", 
			nProcessors, shuffles_ML, IOTime, alnTime, gatherTime, fitTime);
}


int StatSig::FindStringIndex(char* SeqName, SeqInfo* dbSeqs, int dbSeqNum)
{
	if (SeqName == NULL)
	{
		printf("Error : name checked is empty !");
		return -1;
	}
	if (dbSeqNum <= 0) {
		printf("Error : database name is empty !");
		return -1;
	}
	for (int i = 0; i < TotalSeqs; i++) 
	{
		if (strstr (dbSeqs[i].name, SeqName) != NULL)
		{
			 return i;
		}
	}
	return -1;
}

void StatSig::WritePSSToFile (CompRec* PSSInfo, int num, char* PvalueScvFile)
{
		FILE* pEvScv = fopen(PvalueScvFile, "w");
		if(pEvScv == NULL){
			printf("Error: cannot open file %s \n", PvalueScvFile);
			exit(0);
		}
		
		if (PSSInfo == NULL)
		{
			printf("Error: no data stored to file!  \n");
			exit(0);
		}
		
		if (num < 1)
		{
			printf("Error: no data stored to file!  \n");
			return;
		}

		//fprintf(pEvScv,"Seq1No,Seq1name,Seq2No,Seq2name,Len, AlignLen, Topolog,Homolog,Lambda,K,Score,Pvalue\n");
		fprintf(pEvScv,"Seq1No,Seq1name,Seq2No,Seq2name,\
				Topolog,Homolog,Lambda,K,Score,Pvalue\n");
		for (int i = 0 ; i < num; i++){
			fprintf(pEvScv,"%d,%s,%d,%s, %d,%d,%e,%e,%e,%e\n",
			//fprintf(pEvScv,"%d,%s,%d,%s,%d,%d,%d,%d,%e,%e,%e,%e\n",
			PSSInfo[i].seq1,
			PSSInfo[i].name1,
			PSSInfo[i].seq2,
			PSSInfo[i].name2,
			
			//PSSInfo[i].len,
			//PSSInfo[i].alignlen,
			
			PSSInfo[i].topolog,
			PSSInfo[i].homolog,
			PSSInfo[i].lambda,
			PSSInfo[i].K,
			PSSInfo[i].score,
			PSSInfo[i].Pvalue			
			);
		}
		
		fclose(pEvScv);
		
		// release memory
		if (PSSInfo)
		{
			free(PSSInfo);
			PSSInfo = NULL;
		}
}

void StatSig::ReadPSSMFromFile(char* PSSMFilename, int queryLen)
{
	
	if (queryLen <= 0)
	{
		ERROR_OUT("Error:the length of qurey sequence is wrong!\n");
	} else if (queryLen > MaxSeqLen) {
		ERROR_OUT("Error:the length of qurey sequence is too long!\n");
	}
		
	
	FILE* fp_PSSM = fopen(PSSMFilename,"r");
	
	if(fp_PSSM == NULL) 
	{
		ERROR_OUT("Error: Cannot open PSSM file!\n");
	}
	
	char ch;
	int dig;
	
	// skip spaces character
	ch = fgetc(fp_PSSM); ch = fgetc(fp_PSSM);
	ch = fgetc(fp_PSSM); ch = fgetc(fp_PSSM);
	
	// skip the first description line
	while(ch != '\n') ch = fgetc(fp_PSSM);
	
	ch = fgetc(fp_PSSM);
	ch = fgetc(fp_PSSM);
	
	// skip space characters before amino acid residue set,
	// i.e., A  R  N ...
	while(ch == ' ') ch = fgetc(fp_PSSM);
  	

	
	set<char> PSSM_Symbol_Set;
	set<char>::iterator it;
	
	char* resSet = (char*) malloc (32 * sizeof(char));
	if (resSet == NULL)
	{
		ERROR_OUT("Error: Memory allocation fails!\n");
	}
	
	// read amino acid residue set (whose size = 20 or 23, currently)
	// insert the first no space character(i.e., first residue symbol)
	// note that we use STL container: set, just want to find non-reduntant residues
	PSSM_Symbol_Set.insert(ch);
	it = PSSM_Symbol_Set.end();
	resSet[0] = ch;
	int setsNum = 1;
	while (it == PSSM_Symbol_Set.end()) 
	{
		ch = fgetc(fp_PSSM);	// read space, skip it
		ch = fgetc(fp_PSSM);	// read space, skip it
		ch = fgetc(fp_PSSM);	// read amino symbol
		
		it = PSSM_Symbol_Set.find(ch);
		
		// if the residue is not found in the previous set, it's new, then insert the set
		// otherwise, break the while loop
		if (it == PSSM_Symbol_Set.end()) {
			PSSM_Symbol_Set.insert(ch);
			resSet[ setsNum++ ] = ch;
		} else {
			break;
		}
	}
	
	// delete reduntant space in residue set
	it = PSSM_Symbol_Set.find(' ');
	while(it != PSSM_Symbol_Set.end())
	{	
		PSSM_Symbol_Set.erase (it);
		it = PSSM_Symbol_Set.find(' ');
	}
	
	// obtain the correct the maximum number of amino acid residues;
	MAX_AMINO_ACIDS = PSSM_Symbol_Set.size();
	if (MAX_AMINO_ACIDS == 23)
	{
		amino_acids_trans 	= amino_acids_trans_23;
		amino_acids			= amino_acids_23;
	} else if (MAX_AMINO_ACIDS == 20){
		amino_acids_trans 	= amino_acids_trans_20;
		amino_acids			= amino_acids_20;
	} else{
	
		ERROR_OUT("Error: the number of acids amino acid residue is wrong!\n");
	}
	
	// guarantee the privious PSSM matrix memory is released.
	if (PSSMatrix){
		free(PSSMatrix);
		PSSMatrix = NULL;
	}
	
	//initialize  PSSM Matrix as zero	
	PSSMatrix = (int*)calloc(queryLen * ( MAX_AMINO_ACIDS ), sizeof(int));
    
	// HERE: +1 ,just for DUMMY_AMINO_ACID
    if (PSSMatrix == NULL)	{		
        ERROR_OUT("Error: Host memory allocation fails!\n"); 
    }


    
	for (int i = 0; i < queryLen; i++)  
	{	
		while(ch != '\n') // skip to valid character until next line.
		{
			ch = fgetc(fp_PSSM);
		}
		
		// read and skip No. of qurey sequence symbol
		fscanf(fp_PSSM,"%d",&dig); 
		
		ch = fgetc(fp_PSSM); //read and skip space
		ch = fgetc(fp_PSSM); //read and skip amino symbol
		
		int temp;					
		for (int j = 0; j < MAX_AMINO_ACIDS; j++) 
		{
			fscanf(fp_PSSM,"%d",&dig); //read a digital number
						
			temp = amino_acids_trans[resSet[j]];
			if (temp == -1)
			{
				ERROR_OUT("Error: Sequence includes invalid symbol\n");
			}
			else
			{
				// coloum of is mino symbol is lexical sort
				PSSMatrix[ temp * queryLen + i  ] = dig;			
			}

		}
			  
	}
	// just for testing
	#if 0
	it = PSSM_Symbol_Set.begin();
	for (int i = 0; i < MAX_AMINO_ACIDS ; i++)
	{
		cout << *it <<",";
		for(int j = 0; j < queryLen; j++)
		{
			cout << PSSMatrix[i * queryLen + j] << ",";
		}
		cout <<endl;
		it++;
	}
	#endif
	
	if (fp_PSSM) fclose(fp_PSSM);
}


void StatSig::getLocalProfileFromSSM(char* query, int queryLen)
{
	if (queryLen == 0)
	{
		ERROR_OUT("Error:the length of qurey sequence is wrong!\n");
	} else if (queryLen > MaxSeqLen) {
		ERROR_OUT("Error:the length of qurey sequence is too long!\n");
	}
	
	// guarantee the privious local profile matrix memory is released.
	if(LocalMatrix) {
		free(LocalMatrix);
		LocalMatrix = NULL;
	}
	//initialize  PSSM Matrix as zero	
	LocalMatrix = (int*)calloc(queryLen * ( MAX_AMINO_ACIDS ), sizeof(int));
    
    if (LocalMatrix == NULL)	{		
        ERROR_OUT("Error: Host memory allocation fails!\n"); 
    }
	
	char c1, c2;
	char* acids = amino_acids_23;

	for(int i = 0; i < MAX_AMINO_ACIDS; i++)
	{
		c1 = acids[ i ];
		for( int j = 0; j < queryLen; j++)
		{
			c2 = query[ j ];
			
			LocalMatrix[ i * queryLen + j ] = BlmMatrix[ c1 ][ c2 ];
		}
		
	}
	

}


int StatSig::loadClsOfSeqs()
{	

	int Level = 4;		//number of levels of classification
	char tempName[MaxNameLen];
	iniStr(tempName, MaxNameLen);
	
	// allocate memory for array of classification
	dbSeqClf = (int**)malloc(TotalSeqs * sizeof(int*));

	if(dbSeqClf == NULL)
	{
		ERROR_OUT("Error: no memory space for database sequences \n");
	}
	for(int i = 0; i < TotalSeqs; i++)
	{
		dbSeqClf[i] = (int*)malloc(Level * sizeof(int));
		if(dbSeqClf[i] == NULL)
		{
			ERROR_OUT("Error: no memory space for database sequences \n");
		}
	}
		

	int levelCls 	= 0;
	int numSeqs		= 0;

	
	FILE*	fp_clsSeq;
	fp_clsSeq = fopen (seqClsFile, "r");
	if (!fp_clsSeq)
	{
		cout << "Error: opening file '" << seqClsFile <<"' fails!"<<endl;
		MPI::Finalize();
		return 0;
	}
	int readSeq = fscanf(fp_clsSeq, "%s", tempName);
	
	
	char ch;
	int t = 0;
	int temp;
	while(readSeq != EOF)
	{						

		levelCls = 0;
		t = 0;

		fscanf(fp_clsSeq, "%d", &temp);
		dbSeqClf[numSeqs][levelCls++] = temp;
		fscanf(fp_clsSeq, "%c", &ch);

		if(dbSeqClf[numSeqs][levelCls] == EOF || ch == EOF) {
			break;
		} else {
			while(ch == '.') 	//at least one more level of classification
			{
				fscanf(fp_clsSeq, "%d%c", &temp, &ch);
				dbSeqClf[numSeqs][levelCls++] = temp;
				//printf("%d : %d,   %c\n",t++, temp,ch);
				if(t >= 4) break;
			}
		}
		
		readSeq = fscanf(fp_clsSeq, "%s", tempName); // skip sequence name
		//CHECK(tempName)
		numSeqs++;
		#if 10
		if (numSeqs > TotalSeqs) break;
		#endif
		
		//CHECK(numSeqs)
		
	}
	
	return 0;
}


/* 
	read a file containing alphabet and substitution weights 
	note that this function works only for triangle form blosum matrix as follows:
	 ARNDCQEGHILKMFPSTWYVBZX
	 6
	-2  8
	-2 -1  8
	-3 -2  2  9
	....
	if your input matrix is in a different way, please change this function accordingly
*/
void StatSig::getMatrixFromFile(char* BlosumFilename,	int matrix[128][128])
{

	char  alph[100 + 2];		/* alphabet */
	int   size;					/* size of alphabet */
	char  cline[100 + 2];		/* a line of a substitution file */
	int match, mismh, ms;

	FILE* pSubMatFile = fopen(BlosumFilename, "r");
	if(pSubMatFile == NULL){
		fprintf(stderr, "Opening file (%s) failed\n", BlosumFilename);
		exit(0);
	}

	while ( fgets(cline, 100, pSubMatFile) != NULL )
	{ 
		sscanf(cline, "%s", alph);
		if ( alph[0] != '#' )  break;
	}
	size = strlen(alph);		
	for ( int i5 = 0; i5 < 128 ; i5++ )
	{
		for ( int j5 = 0; j5 < 128 ; j5++ )
		{
			matrix[i5][j5] = 0;
		}
	}

	match = mismh = 0;
	for ( int i5 = 0; i5 < size ; i5++ )
	{
		for ( int j5 = 0; j5 <= i5 ; j5++ )
		{ 
			(void) fscanf(pSubMatFile, "%d", &ms);
			matrix[alph[i5]][alph[j5]] = matrix[alph[j5]][alph[i5]] = ms;
			
			if ( ms > match ) match = ms;
			if ( ms < mismh ) mismh = ms;
		}
	}
	if (pSubMatFile) fclose(pSubMatFile);
}

//for quick sort
int	StatSig::compar_ascent_Pvalue(const void* va, const void* vb)
{
	const CompRec* a = (const CompRec*)va;
	const CompRec* b = (const CompRec*)vb;

	if(a->Pvalue > b->Pvalue)	return 1;
	if(a->Pvalue < b->Pvalue) return -1;

	return 0;
}
int StatSig::compar_descent_Pvalue(const void* va, const void* vb)
{
    const CompRec* a = (const CompRec*)va;
    const CompRec* b = (const CompRec*)vb;

    if(a->Pvalue > b->Pvalue) return -1;
    if(a->Pvalue < b->Pvalue) return 1;

    return 0;
}

void StatSig::writePSSToFile (CompRec* PSSInfo, int num, char* EvalueScvFile)
{
		FILE* pEvScv = fopen(EvalueScvFile, "w");
		if(pEvScv == NULL){
			printf("Error: cannot open file %s \n", EvalueScvFile);
			exit(0);
		}
		
		if (PSSInfo == NULL)
		{
			printf("Error: no data stored to file!  \n");
			exit(0);
		}
		
		if (num < 1)
		{
			printf("Error: no data stored to file!  \n");
			return;
		}

		fprintf(pEvScv,"Seq1No,Seq1name,Seq2No,Seq2name,\
				Topolog,Homolog,Lambda,K,Score,Evalue\n");
		for (int i = 0 ; i < num; i++){
			fprintf(pEvScv,"%d,%s,%d,%s, %d,%d,%e,%e,%e,%e\n",
			PSSInfo[i].seq1,
			PSSInfo[i].name1,
			PSSInfo[i].seq2,
			PSSInfo[i].name2,
			
			
			PSSInfo[i].topolog,
			PSSInfo[i].homolog,
			PSSInfo[i].lambda,
			PSSInfo[i].K,
			PSSInfo[i].score,
			PSSInfo[i].Pvalue			
			);
		}
		
		fclose(pEvScv);
		
}


