//Modified by Ankit Agrawal

/* A Global Alignment Program (GAP3) for Comparing Sequences:

Copyright (c) 2002 Xiaoqiu Huang
All Rights Reserved.  E-mail: xqhuang@cs.iastate.edu

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice, this paragraph and the following three paragraphs
appear in all copies.

Permission to incorporate this software into commercial products may be
obtained from Iowa State University Research Foundation, Inc.

	      Xiaoqiu Huang
	      Department of Computer Science
	      Iowa State University
	      Ames, IA 50011
	      E-mail: xqhuang@cs.iastate.edu
	      WWW: http://www.cs.iastate.edu/~xqhuang

     Proper attribution of the author as the source of the software would
     be appreciated:
        Huang, X. and Chao, K.-M. (2002)
	In preparation.

    The GAP program computes a global alignment of two sequences
    without penalizing terminal gaps. It delivers the alignment in
    linear space, so long sequences can be aligned. 

    Users supply scoring parameters. In the simplest form, users just
    provide 3 integers: ms, q and r, where ms is the score of a mismatch
    and the score of an i-symbol indel is -(q + r * i). Each match
    automatically receives score 10. This simple scoring scheme may be
    used for DNA sequences. NOTE: all scores are integers.

    In general, users can define an alphabet of characters appearing
    in the sequences and a matrix that gives the substitution score
    for each pair of symbols in the alphabet. The 127 ASCII characters
    are eligible. The alphabet and matrix are given in a file, where
    the first line lists the characters in the alphabet and the lower
    triangle of the matrix comes next. An example file looks as follows:

    ARNDC	       
     13
    -15  19
    -10 -22  11
    -20 -10 -20  18
    -10 -20 -10 -20  12

    Here the -22 at position (3,2) is the score of replacing N by R.
    This general scoring scheme is useful for protein sequences where the
    set of protein characters and Dayhoff matrix are specified in the file.

    The GAP program is written in C and runs under Unix systems on
    Sun workstations and under DOS systems on PCs.
    We think that the program is portable to many machines.

    Sequences to be analyzed are stored in separate files.
    Sequences must be in FASTA format. The first line begins with the symbol '>'
    followed by the name of the sequence. The sequence is on the remaining lines.
    Protein sequences must be in upper case.
    DNA sequences could be in upper or lower case.
    A sample sequence file is shown below.

>DNA sequence
GAATTCTAATCTCCCTCTCAACCCTACAGTCACCCATTTGGTATATTAAA
GATGTGTTGTCTACTGTCTAGTATCCCTCAAGTAGTGTCAGGAATTAGTC
ATTTAAATAGTCTGCAAGCCAGGAGTGGTGGCTCATGTCTGTAATTCCAG
CACTGGAGAGGTAGAAGTG

    To find the best alignment of two sequences in files A and B,
    use a command of form

	   gap  A  B  gs  ms  q  r > result

    where gap is the name of the object code, gs is the minimum length
    of any gap in the short sequence receiving a constant gap penalty,
    ms is a negative integer specifying mismatch weight, q and r are
    non-negative integers specifying gap-open and gap-extend penalties,
    respectively. Output alignment is saved in the file "result".

    For using a scoring matrix defined in file S, use a command of form

	   gap  A  B  gs  S  q  r > result

    Note that ms is replaced by the file S.
*/

#include   <stdio.h>
#include 	<stdlib.h>
#include	<string.h>
#include 	"gap3PSSM.comp.h"
#include 	"generic.h"
#include <iostream>
using namespace std;

#define  MATCHSC    10   /* match score */
#define  LINELEN    60   /* lenght of an output line */
static int numrows = 30; /* max number of rows in the basis */

static int match, mismh;		/* max and min substitution weights */
static int isdna;                       /* 1, DNA; 0, protein */
static int v[5000][128];	/* substitution scores */
static int  q, r;       /* gap penalties */
static int  qr;         /* qr = q + r */
static int  pay;	/* constant-cost for gaps */
static int  neginf = -10000;	/* negative infinity */
static long  totalmemreq = 0;   /* total amount of memory in bytes */
static int   maxM, maxN;        /* maximum sequence lengths */

static int *CC, *DD;	/* saving matrix scores */
static int *RR, *SS; 	/* saving start-points */
static int  *S;		/* saving operations for diff */
static int *HH;		/* for long deletion gaps in forward pass */
static int *XX;		/* for long deletion gaps in reverse pass */
static short *MD;	/* a list of flags for major differences */

struct TraceType
   { char c, d, e, h; }; /* a traceback record */
   static struct TraceType **tracem; /* traceback information */


/* The following definitions are for function diff() */

static int  diff();
static int  zero = 0;				/* int type zero        */

#define gap(k)  ((k) <= 0 ? 0 : q+r*(k))	/* k-symbol indel score */

static int *sapp;				/* Current script append ptr */
static int  last;				/* Last script op appended */
static short *mdptr;				/* pointer to current major difference flag */
static short premd;				/* last major difference flag */

static int no_mat; 				/* number of matches */ 
static int no_mis; 				/* number of mismatches */ 
static int al_len; 				/* length of alignment */
static int no_regs; 				/* number of similar regions */
static int sti, stj, edi, edj;  /* start and end points of an alignment */
static int mscore;		/* max score of alignments */
						/* Append "Delete k" op */
#define DEL(k, md)			\
{ if ( ! md ) al_len += (k);		\
  if ( last < 0 && premd == md )	\
    last = sapp[-1] -= (k);		\
  else					\
   { last = *sapp++ = -(k);		\
     premd = *mdptr++ = md;		\
   }                                    \
}
						/* Append "Insert k" op */
#define INS(k, md)			\
{ if ( ! md ) al_len += (k);		\
  if ( last > 0 && premd == md )	\
    last = sapp[-1] += (k);		\
  else					\
   { last = *sapp++ = (k);		\
     premd = *mdptr++ = md;		\
   }                                    \
}
						/* Append "Replace" op */
#define REP 				\
{ if ( premd ) no_regs++;               \
  last = *sapp++ = 0; 			\
  premd = *mdptr++ = 0;                 \
  al_len++;				\
}

void AllocateArrays(int M, int N)
{ int  score;   		/* the max score in LIST */
  int  i, j;			/* row and column indices */
  //void *ckalloc();		/* space-allocating function */
	
	    maxM = M;
	    maxN = N;
	    isdna = 0;
	    /* allocate space for all vectors */
	    j = (N + 1) * sizeof(int);
	    CC = ( int * ) ckalloc(j);
	    DD = ( int * ) ckalloc(j);
	    RR = ( int * ) ckalloc(j);
	    SS = ( int * ) ckalloc(j);
	    HH = ( int * ) ckalloc(j);
	    XX = ( int * ) ckalloc(j);
	    i = (M + 1) * sizeof(int);
	    S = ( int * ) ckalloc(i + j);
	    MD = ( short * ) ckalloc( (M + N + 2) * sizeof(short) );
	    i = (numrows + 1) * sizeof(struct TraceType *);
	    tracem = ( struct TraceType ** ) ckalloc(i);
	    j = (N + 1) * sizeof(struct TraceType);
	    for ( i = 0; i <= numrows; i++ )
	      tracem[i] = ( struct TraceType * ) ckalloc(j);
}
void FreeArrays()
{ 
	/* free memory space for all vectors */
	if(CC) {free(CC); CC = NULL;}
	if(DD) {free(DD); DD = NULL;}
	if(RR) {free(RR); RR = NULL;}
	if(SS) {free(SS); SS = NULL;}	
	if(HH) {free(HH); HH = NULL;}
	if(XX) {free(XX); XX = NULL;}
	if(MD) {free(MD); MD = NULL;}
			
	int i;
	for ( i = 0; i <= numrows; i++ )
	{
	  if(tracem[i]) {free (tracem[i]); }
	}
	
	if(tracem) {free (tracem); tracem = NULL;}
}

static int vv[5000][128]; //PSSM
static char PSSMaa[20];

void extractPSSMFromFile(int vv[][128], char *filename, int len)
{
	char ch;
	int dig;
	
	FILE *fp =fopen(filename,"r");
	
	if(fp==NULL) {
		printf("\nCannot open PSSM file %s",filename);
		exit(0);
	}

	ch = fgetc(fp);ch = fgetc(fp);ch = fgetc(fp);ch = fgetc(fp);
	while(ch!='\n'&& ch != EOF)
		ch=fgetc(fp);
	ch = fgetc(fp);ch = fgetc(fp);
	while(ch==' '&& ch != EOF)
		ch=fgetc(fp);
		
	PSSMaa[0]=ch;
	int i,ii;
	for (i=1;i<20;i++) {
		ch = fgetc(fp);ch = fgetc(fp);ch = fgetc(fp);
		PSSMaa[i]=ch;
		//printf("%c",PSSMaa[i]);
	}
	
	for (ii=0;ii<len;ii++) 
	{
		
		for (i=0;i<128;i++)
			vv[ii][i]=0;

		while(ch!= '\n' && ch != EOF)
			ch=fgetc(fp);
		

		fscanf(fp,"%d",&dig);

		ch = fgetc(fp);ch = fgetc(fp);
		
		for (i=0;i<20;i++) {
			fscanf(fp,"%d",&dig);
			
			vv[ii][PSSMaa[i]]=dig;
		}

	}

	if(fp) fclose(fp);

}


		
void GetPSSM( char* pssmFile,		/* substition table */
			int   qq,		/* gap open */
			int   rr,		/* gap extension */
			int   ppay,		/* difference block */
			int  M, int N)      /* their lengths */
{

	AllocateArrays(M, N);

	extractPSSMFromFile(vv, pssmFile, M);

	q = qq;
	r = rr;
	qr = q + r;
	pay = ppay;
	match = mismh = 0;
	int  i, j;			/* row and column indices */
	int  ms;
	for ( i = 0; i < M; i++ )
	  for ( j = 0; j < 128; j++ )
	   { ms = v[i][j] = vv[i][j];
	     if ( ms > match ) match = ms;
	     if ( ms < mismh ) mismh = ms;
	   }
	
}

static int v5[128][128];	//SM

void GetSM(
			char smFile[],		/* substition table */
			int   qq,			/* gap open */
			int   rr,			/* gap extension */
			int   ppay,			/* difference block */
			int  M, int N       /* their lengths */
			)
{
	AllocateArrays(M, N);

	q = qq;
	r = rr;
	qr = q + r;
	pay = ppay;
	match = mismh = 0;

  char  alph[100+ 2], *s;		/* alphabet */
  int   size;				/* size of alphabet */
  char  *name;				/* temporary pointer */
  char  rdline[100+ 2];		/* a line */
  char  sfname[100+ 2];		/* name of a substitution file */
  char  cline[100+ 2];		/* a line of a substitution file */
  char  oflag;		/* flag for a parameter set */	
	
int i5,k5,j5,ms;
	    for ( i5 = 0; i5 < 128 ; i5++ )
	      for ( j5 = 0; j5 < 128 ; j5++ )
	            v5[i5][j5] = 0;

	  { /* read a file containing alphabet and substitution weights */
	    FILE* Sp = fopen(smFile, "r");
	    while ( fgets(cline, 100, Sp) != NULL )
	     { sscanf(cline, "%s", alph);
	       if ( alph[0] != '#' )  break;
	     }
	    size = strlen(alph);
	    match = mismh = 0;
	    for ( i5 = 0; i5 < size ; i5++ )
	     for ( j5 = 0; j5 <= i5 ; j5++ )
		{ fscanf(Sp, "%d", &ms);
		  v5[alph[i5]][alph[j5]] = v5[alph[j5]][alph[i5]] = ms;
		  if ( ms > match ) match = ms;
		  if ( ms < mismh ) mismh = ms;
		}
	    fclose(Sp);
	  }
}

void SM2PSSM(char seq[],int M)
{
	int i,j;

	for ( i = 0; i < M; i++ )
	  for ( j = 0; j < 128; j++ )
	   { 
		v[i][j] = v5[seq[i+1]][j];
	   }

}

int UseGAP3(char A[], char B[],          /* sequences */
			int  M, int N                 /* their lengths */
			)
{ int  score;   		/* the max score in LIST */
  int  i, j;			/* row and column indices */
  int  ms;
  int  tmp;
	
//extractPSSMFromFile(vv, pssmFile, M);	//being called in GetPSSM()

//   Below part copied in GetPSSM() and thus commented
//	q = qq;
//	r = rr;
//	qr = q + r;
//	pay = ppay;
//	match = mismh = 0;
//	for ( i = 0; i < M; i++ )
//	  for ( j = 0; j < 128; j++ )
//	   { ms = v[i][j] = vv[i][j];
//	     if ( ms > match ) match = ms;
//	     if ( ms < mismh ) mismh = ms;
//	   }


	if ( M == 0 && N == 0 )
	   fatal("Both sequences are empty");

	if ( pay < 1 )
	   fatal("Penalty for a major difference is a positive integer");

	if ( q < 0 )
	   fatal("The gap-open penalty is a nonnegative integer");

	if ( r < 0 )
	   fatal("The gap-extend penalty is a nonnegative integer");

  	if ( M > maxM || N > maxN )
	  fatal("Sequences are too long");
/*
(void) printf("Max Match  Min Mismatch  Gap-Open Penalty  Gap-Ext. Pen.  Diff. Block Pen.\n");
(void) printf("   %d          %d              %d               %d              %d\n\n",
                                       match, mismh, q,  r, pay);
	    (void) printf("Upper Sequence: %s\n", dhead1);
	    (void) printf("      Length: %d\n", M);
	    (void) printf("Lower Sequence: %s\n", dhead2);
	    (void) printf("      Length: %d\n\n", N);  */

	    ScoreOnly(A, B, M, N, 0, -q, -pay, 0, -q, -pay);
            return mscore;
	    /*
	    mscore = sti = stj = edi = edj = 0;
	    FindEndPoints(A, B, M, N, 0, -q, -pay, 0, -q, -pay);
	    score = 0;
            sapp = S;
            last = 0;
	    mdptr = MD;
	    premd = 1;
            al_len = no_mat = no_mis = no_regs = 0;
	    score = diff2(A+sti, B+stj, edi-sti, edj-stj, 0, -q, -pay, 0, -q, -pay);
(void) printf("      Best Alignment with Constant Penalty for Difference Blocks\n");
(void) printf("      Similarity Score: %d\n",score);
(void) printf("      Cost-only Score:  %d\n",mscore);
tmp = al_len > 0 ? al_len : 1;
(void) printf("      Percent Identity: %d%%\n", (100*no_mat)/tmp);
(void) printf("      Number of Matches: %d\n", no_mat);
(void) printf("      Number of Mismatches: %d\n", no_mis);
(void) printf("      Length of Gaps: %d\n", al_len-no_mat-no_mis);
(void) printf("      Number of Similar Regions: %d\n", no_regs);
(void) printf("      Length of Similar Regions: %d\n", al_len);
if ( mscore != score ) fatal("Different scores between score-only and alignment");
(void) printf("      Begins at (%d, %d) and Ends at (%d, %d)\n",
                                    sti + 1, stj + 1, edi, edj);
	    display(A+sti, B+stj, edi-sti, edj-stj, S, MD , sti + 1 , stj + 1);
            return score;  */
}

void ScoreOnly(
		char* A, char* B, 
		int M, int N, 
		int c00, int d00, 
		int h00, int cmn, 
		int dmn, int hmn
		)
{ int   i, j;
  int   c, e, d, s, h;
  int   t, *va;
  int   g, y, tmp;
  int    maxs;

  maxs = 0;		/* Forward phase */
  CC[0] = c = c00; /* Compute C(M/2,k), D(M/2,k) & H(M/2,k) for all k */
  e = c00 - q;
  g = h00 > ( y = c00 - pay ) ? h00 : y;
  for (j = 1; j <= N; j++)
   { e -= r;
     c = e > g ? e : g;
     if ( c < 0 )  c = 0;
     CC[j] = c;
     DD[j] = c - q;
     HH[j] = g;
   }
  t = d00 > (y = c00 - q) ? d00 : y;
  for (i = 1; i <= M; i++)
    { s = CC[0];
      t -= r;
      c = t > g ? t : g;
      if ( c < 0 )  c = 0;
      CC[0] = c;
      e = c - q;
      h = g;
      //va = v[A[i]];
	  va = v[i];
      for (j = 1; j <= N; j++)
        { if ((tmp = c - qr) > (e = e - r)) e = tmp;
	  if ( c < (tmp = CC[j] ) ) c = tmp;
	  if ( (c = c - pay) < (y = HH[j]) ) c = y;
	  if ( h < c ) h = c;
          if ((c = tmp - qr) > (d = DD[j] - r)) d = c;
	  c = s+va[B[j]];
          if ( c < d ) c = d;
          if ( c < e ) c = e;
	  if ( c < h ) c = h;
	  if ( c < 0 ) c = 0;
	  if ( maxs < c ) 
	   { maxs = c; edi = i; edj = j; }
          s = tmp;
          CC[j] = c;
          DD[j] = d;
          HH[j] = h;
        }
    }
  DD[0] = t;
  HH[0] = g;
  mscore = maxs;
}

void FindEndPoints(
		char* A, char* B, 
		int M, int N, 
		int c00, int d00, 
		int h00, int cmn, 
		int dmn, int hmn
		)
{ int   i, j;
  int   c, e, d, s, h;
  int   t, *va;
  int   g, y, tmp;
  int    maxs;

  maxs = 0;		/* Forward phase */
  CC[0] = c = c00; /* Compute C(M/2,k), D(M/2,k) & H(M/2,k) for all k */
  e = c00 - q;
  g = h00 > ( y = c00 - pay ) ? h00 : y;
  for (j = 1; j <= N; j++)
   { e -= r;
     c = e > g ? e : g;
     if ( c < 0 )  c = 0;
     CC[j] = c;
     DD[j] = c - q;
     HH[j] = g;
   }
  t = d00 > (y = c00 - q) ? d00 : y;
  for (i = 1; i <= M; i++)
    { s = CC[0];
      t -= r;
      c = t > g ? t : g;
      if ( c < 0 )  c = 0;
      CC[0] = c;
      e = c - q;
      h = g;
      //va = v[A[i]];
	  va = v[i-1];
      for (j = 1; j <= N; j++)
        { if ((tmp = c - qr) > (e = e - r)) e = tmp;
	  if ( c < (tmp = CC[j] ) ) c = tmp;
	  if ( (c = c - pay) < (y = HH[j]) ) c = y;
	  if ( h < c ) h = c;
          if ((c = tmp - qr) > (d = DD[j] - r)) d = c;
	  c = s+va[B[j]];
          if ( c < d ) c = d;
          if ( c < e ) c = e;
	  if ( c < h ) c = h;
	  if ( c < 0 ) c = 0;
	  if ( maxs < c ) 
	   { maxs = c; edi = i; edj = j; }
          s = tmp;
          CC[j] = c;
          DD[j] = d;
          HH[j] = h;
        }
    }
  DD[0] = t;
  HH[0] = g;
  if ( edi == 0 ) return;
  M = edi; N = edj;
  mscore = maxs;

  maxs = 0;
  RR[N] = c = cmn; /* Reverse phase */
  e = cmn - q; /* Compute R(M/2,k), S(M/2,k) & X(M/2,k) for all k */
  g = hmn > ( y = cmn - pay ) ? hmn : y;
  for (j = N-1; j >= 0; j--)
   { e -= r;
     c = e > g ? e : g;
     if ( c < 0 )  c = 0;
     RR[j] = c;
     SS[j] = c - q;
     XX[j] = g;
   }
  t = dmn > (y = cmn - q) ? dmn : y;
  for (i = M-1; i >= 0; i--)
    { s = RR[N];
      t -= r;
      c = t > g ? t : g;
      if ( c < 0 )  c = 0;
      RR[N] = c;
      e = c - q;
      h = g;
      //va = v[A[i+1]];
	  va = v[i];
      for (j = N-1; j >= 0; j--)
        { if ( (tmp = c - qr) > (e = e - r) ) e = tmp;
	  if ( c < (tmp = RR[j] ) ) c = tmp;
	  if ( (c = c - pay) < (y = XX[j]) ) c = y;
	  if ( h < c ) h = c;
          if ((c = tmp - qr) > (d = SS[j] - r)) d = c;
	  c =  s+va[B[j+1]];
          if ( c < d ) c = d;
          if ( c < e ) c = e;
	  if ( c < h ) c = h;
	  if ( c < 0 ) c = 0;
	  if ( maxs < c )
	   { maxs = c; sti = i; stj = j; }
          s = tmp;
          RR[j] = c;
          SS[j] = d;
          XX[j] = h;
        }
    }
  SS[N] = t;
  XX[N] = g;
}

/* diff2(A,B,M,N,c00,d00,h00,cmn,dmn,hmn) returns the score of an optimum
   conversion between A[1..M] and B[1..N] such that the upper left and
   lower right corners are initialized to the given values:
   C(0,0) = c00, D(0,0) = d00, H(0,0) = h00,
   C(m,n) = cmn, D(m,n) = dmn, H(m,n) = hmn. */

int diff2(
		char* A, char* B, 
		int M, int N, 
		int c00, int d00, int h00, 
		int cmn, int dmn, int hmn
		)

{ int  midi, midj, type;	/* Midpoint, type, and cost */
  int  midc;
  int  cmid, dmid, hmid;	/* C, D & H values at midpoint */

{ register int   i, j;
  register int   c, e, d, s, h;
           int   t, *va;
	   int   g, y, tmp;
	   struct TraceType *trow; /* one row of tracem */
	   struct TraceType *ten;  /* pointer to one entry of tracem */
	   char             letc;  /* a letter to indicate matrix type */


/* Boundary cases: M <= numrows */

if ( M <= numrows )
{ trow = tracem[M]; ten = trow + N;
  RR[N] = c = cmn; ten->c = 'c';
  e = c - q; ten->e = 'c';
  if ( dmn > e )
   { t = dmn; ten->d = 'd'; }
  else
   { t = e; ten->d = 'c'; }
  if ( hmn > ( y = c - pay ) )
   { g = hmn; ten->h = 'h'; }
  else
   { g = y; ten->h = 'c'; }
  for (j = N-1; j >= 0; j--)
   { ten = trow + j;
     e -= r; ten->e = 'e';
     if ( e > g )
      { RR[j] = c = e; ten->c = 'e'; }
     else
      { RR[j] = c = g; ten->c = 'h'; }
     SS[j] = c + neginf; ten->d = 'c';
     XX[j] = g; ten->h = 'g';
   }
  for (i = M-1; i >= 0; i--)
    { trow = tracem[i]; ten = trow + N;
      s = RR[N];
      t -= r; ten->d = 'd';
      if ( t > g )
       { RR[N] = c = t; ten->c = 'd'; }
      else
       { RR[N] = c = g; ten->c = 'h'; }
      e = c + neginf; ten->e = 'c';
      h = g; ten->h = 'h';
      //va = v[A[i+1]];
	  va = v[i];		
      for (j = N-1; j >= 0; j--)
        { ten = trow + j;
          if ( (tmp = c - qr) > (e = e - r) ) { e = tmp; ten->e = 'c'; }
	  else ten->e = 'e';
	  if ( c < (tmp = RR[j] ) ) { c = tmp; ten->h = 'x'; }
	  else ten->h = 'y';
	  if ( (c = c - pay) < (y = XX[j]) ) { c = y; ten->h = 'h'; }
	  if ( h < c ) h = c;
	  else  ten->h = 'g';
          if ((c = tmp - qr) > (d = SS[j] - r)) { d = c; ten->d = 'c'; }
	  else ten->d = 'd';
	  c =  s+va[B[j+1]]; ten->c = 'c';
          if ( c < d ) { c = d; ten->c = 'd'; }
          if ( c < e ) { c = e; ten->c = 'e'; }
	  if ( c < h ) { c = h; ten->c = 'h'; }
          s = tmp;
          RR[j] = c;
          SS[j] = d;
          XX[j] = h;
        }
    }
  SS[N] = t;
  XX[N] = g;

  midc = RR[0] + c00; letc = 'c';
  if ( midc < (tmp = XX[0] + h00 + pay) )
    { midc = tmp; letc = 'h'; }
  if ( midc < (tmp = SS[0] + d00 + q) )
    { midc = tmp; letc = 'd'; }
  i = j = 0;
  while ( i < M || j < N )
   { switch ( letc )
{  case 'c' : if ( (letc = tracem[i][j].c ) == 'c' )
                { if ( i >= M || j >= N )
		    fatal("Row/column index exceeds the limit");
		  i++; j++; REP
	          if ( A[i] == B[j] || isdna && v[i-1][B[j]] == MATCHSC )
	            no_mat += 1;
	          else
	            no_mis += 1;
		}
	      break;
   case 'd' : letc = tracem[i][j].d;
              if ( i >= M ) fatal("D: row index exceeds the limit");
	      i++; DEL(1, 0)
	      break;
   case 'e' : letc = tracem[i][j].e;
              if ( j >= N ) fatal("E: column index exceeds the limit");
	      j++; INS(1, 0)
	      break;
   case 'h' : letc = tracem[i][j].h;
              if ( letc == 'h' || letc == 'x' )
                { if ( i >= M ) fatal("Row index exceeds the limit");
	          i++; DEL(1, 1)
		  if ( letc == 'x' ) letc = 'c';
		}
              else
                { if ( j >= N ) fatal("Column index exceeds the limit");
	          j++; INS(1, 1)
		  if ( letc == 'g' ) letc = 'h';
		  else
		    if ( letc == 'y' ) letc = 'c';
		    else
                     fatal("Wrong traceback letter");
		}
              break;
    default : printf("Letter %c\n", letc);
              fatal("Wrong traceback letter");
}
   }
 return midc;
}

/* Divide: Find optimum midpoint (midi,midj) of cost midc */

  midi = M/2;	   /* Forward phase */
  CC[0] = c = c00; /* Compute C(M/2,k), D(M/2,k) & H(M/2,k) for all k */
  e = c00 - q;
  g = h00 > ( y = c00 - pay ) ? h00 : y;
  for (j = 1; j <= N; j++)
   { e -= r;
     CC[j] = c = e > g ? e : g;
     DD[j] = c - q;
     HH[j] = g;
   }
  t = d00 > (y = c00 - q) ? d00 : y;
  for (i = 1; i <= midi; i++)
    { s = CC[0];
      t -= r;
      CC[0] = c = t > g ? t : g;
      e = c - q;
      h = g;
      //va = v[A[i]];
	  va = v[i-1];
      for (j = 1; j <= N; j++)
        { if ((tmp = c - qr) > (e = e - r)) e = tmp;
	  if ( c < (tmp = CC[j] ) ) c = tmp;
	  if ( (c = c - pay) < (y = HH[j]) ) c = y;
	  if ( h < c ) h = c;
          if ((c = tmp - qr) > (d = DD[j] - r)) d = c;
	  c = s+va[B[j]];
          if ( c < d ) c = d;
          if ( c < e ) c = e;
	  if ( c < h ) c = h;
          s = tmp;
          CC[j] = c;
          DD[j] = d;
          HH[j] = h;
        }
    }
  DD[0] = t;
  HH[0] = g;

  RR[N] = c = cmn; /* Reverse phase */
  e = cmn - q; /* Compute R(M/2,k), S(M/2,k) & X(M/2,k) for all k */
  g = hmn > ( y = cmn - pay ) ? hmn : y;
  for (j = N-1; j >= 0; j--)
   { e -= r;
     RR[j] = c = e > g ? e : g;
     SS[j] = c - q;
     XX[j] = g;
   }
  t = dmn > (y = cmn - q) ? dmn : y;
  for (i = M-1; i >= midi; i--)
    { s = RR[N];
      t -= r;
      RR[N] = c = t > g ? t : g;
      e = c - q;
      h = g;
      //va = v[A[i+1]];
	  va = v[i];
      for (j = N-1; j >= 0; j--)
        { if ( (tmp = c - qr) > (e = e - r) ) e = tmp;
	  if ( c < (tmp = RR[j] ) ) c = tmp;
	  if ( (c = c - pay) < (y = XX[j]) ) c = y;
	  if ( h < c ) h = c;
          if ((c = tmp - qr) > (d = SS[j] - r)) d = c;
	  c =  s+va[B[j+1]];
          if ( c < d ) c = d;
          if ( c < e ) c = e;
	  if ( c < h ) c = h;
          s = tmp;
          RR[j] = c;
          SS[j] = d;
          XX[j] = h;
        }
    }
  SS[N] = t;
  XX[N] = g;

  midc = HH[N]+XX[N]+pay;       /* Find optimal midpoint */
  midj = N;
  type = 3;
  for ( j = N; j >= 0; j-- )
   { if ( (c = HH[j] + XX[j] + pay) > midc )
       { midc = c;
         midj = j;
         type = 3;
       }
     if ((c = DD[j] + SS[j] + q) > midc)
       { midc = c;
         midj = j;
         type = 2;
       }
     if ( (c = CC[j] + RR[j]) > midc )
        { midc = c;
          midj = j;
          type = 1;
        }
   }
}

/* Conquer: recursively around midpoint */

  cmid = CC[midj];
  dmid = DD[midj];
  hmid = HH[midj];
  (void) diff2(A,B,midi,midj,c00,d00,h00,RR[midj],SS[midj],XX[midj]);
  (void) diff2(A+midi,B+midj,M-midi,N-midj,cmid,dmid,hmid,cmn,dmn,hmn);
  return midc;
}

/* Alignment display routine */

static char ALINE[LINELEN+1], BLINE[LINELEN+1], CLINE[LINELEN+1];

void display(
		char A[], char B[],		/* sequences */
		int M, int N,		/* their lengths */
		int S[],		/* list of operations */
		short MD[],		/* list of major difference flags */
		int AP, int BP		/* start positions in sequences */
		)
{ register char *a, *b, *c;
  register int   i,  j, op;
           int   lines, ap, bp;
  int gsize;     /* current gap length */
  int gcount;    /* length of a gap removed */
  int flag;      /* 1, print */
  short cmd;	 /* current major difference flag */

  i = j = op = lines = gsize = gcount = 0;
  ap = AP;
  bp = BP;
  a = ALINE;
  b = BLINE;
  c = CLINE;
  while (i < M || j < N)
    { if (op == 0 && *S == 0)
        { op = *S++;
	  cmd = *MD++;
          *a = A[++i];
          *b = B[++j];
	  *c++ = (*a == *b || isdna && v[i-1][*b] == MATCHSC ) ? '|' : ' ';
	  a++;
	  b++;
	  gsize = 0;
        }
      else
        { if (op == 0)
           { op = *S++; gsize = 0; cmd = *MD++; }
          if (op > 0)
            { *a++ = ' ';
              *b++ = B[++j];
              op--;
            }
          else
            { *a++ = A[++i];
              *b++ = ' ';
              op++;
            }
	  if ( cmd )
	   { *c++ = '+'; gsize++; }
	  else
             *c++ = '-';
        }
      if (a >= ALINE+LINELEN || i >= M && j >= N)
        { *a = *b = *c = '\0';
	  flag = 0;
	  if ( gsize < 2 * LINELEN )
	      flag = 1;
	  else
	   if ( op >= 0 && op < LINELEN || op < 0 && -op < LINELEN )
	    { if ( gcount > 0 )
	       { (void) printf("\n              ~~~ %d bp removed ~~~\n", gcount);
	         gcount = 0;
	       }
	      flag = 1;
	    }
	   else
	     gcount += LINELEN;
	  if ( flag )
	   { (void) printf("\n%7d ",LINELEN*lines);
             for (b = ALINE+10; b <= a; b += 10)
               (void) printf("    .    :");
             if (b <= a+5)
                (void) printf("    .");
             (void) printf("\n%7d %s\n        %s\n%7d %s\n",ap,ALINE,CLINE,bp,BLINE);
	   }
	  lines++;
	  ap = AP + i;
	  bp = BP + j;
          a = ALINE;
          b = BLINE;
          c = CLINE;
        }
    }
}
